/*
 * #%L
 * JRst :: Api
 * 
 * $Id: JRST.java 567 2011-04-27 11:43:27Z echatellier $
 * $HeadURL: http://svn.nuiton.org/svn/jrst/tags/jrst-1.3/jrst/src/main/java/org/nuiton/jrst/JRST.java $
 * %%
 * Copyright (C) 2004 - 2011 CodeLutin, Chatellier Eric
 * %%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as 
 * published by the Free Software Foundation, either version 3 of the 
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Lesser Public License for more details.
 * 
 * You should have received a copy of the GNU General Lesser Public 
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/lgpl-3.0.html>.
 * #L%
 */

package org.nuiton.jrst;

import java.awt.GraphicsDevice;
import java.awt.GraphicsEnvironment;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.Reader;
import java.io.StringReader;
import java.net.URL;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.URIResolver;
import javax.xml.transform.sax.SAXResult;
import javax.xml.transform.sax.SAXSource;
import javax.xml.transform.stream.StreamSource;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.fop.apps.FOUserAgent;
import org.apache.fop.apps.Fop;
import org.apache.fop.apps.FopFactory;
import org.apache.fop.apps.MimeConstants;
import org.dom4j.Document;
import org.dom4j.io.OutputFormat;
import org.dom4j.io.XMLWriter;
import org.nuiton.i18n.I18n;
import org.nuiton.i18n.init.ClassPathI18nInitializer;
import org.nuiton.jrst.convertisor.DocUtils2RST;
import org.nuiton.jrst.convertisor.DocUtilsVisitor;
import org.nuiton.util.ApplicationConfig;
import org.nuiton.util.FileUtil;
import org.nuiton.util.Resource;
import org.nuiton.util.StringUtil;
import org.xml.sax.EntityResolver;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;

/**
 * FIXME: 'JRST --help' doesn't work, but 'JRST --help toto' work :( FIXME:
 * 'JRST -c' doesn't work, but 'JRST -c toto'
 * 
 * Created: 3 nov. 06 20:56:00
 *
 * @author poussin
 * @version $Revision: 567 $
 *
 * Last update: $Date: 2011-04-27 13:43:27 +0200 (mer., 27 avril 2011) $
 * by : $Author: echatellier $
 */
public class JRST {

    public enum Overwrite {
        NEVER, IFNEWER, ALLTIME
    }

    /** to use log facility, just put in your code: log.info("..."); */
    protected static Log log = LogFactory.getLog(JRST.class);

    /** XSL Stylesheet to transform RST into HTML. */
    protected static final String rst2xhtml = "/xsl/rst2xhtml.xsl";
    /** XSL Stylesheet to transform RST into Xdoc. */
    protected static final String rst2xdoc = "/xsl/rst2xdoc.xsl";
    /** XSL Stylesheet to transform RST into Docbook. */
    protected static final String rst2docbook = "/xsl/dn2dbk.xsl";

    /** XSL Stylesheet to transform Docbook into xHTML. */
    protected static final String docbook2xhtml = "/docbook/xhtml/docbook.xsl";
    /** XSL Stylesheet to transform Docbook into javahelp. */
    protected static final String docbook2javahelp = "/docbook/javahelp/javahelp.xsl";
    /** XSL Stylesheet to transform Docbook into htmlhelp. */
    protected static final String docbook2htmlhelp = "/docbook/htmlhelp/htmlhelp.xsl";
    
//    /** XSL Stylesheet to transform xml into rst. */
//    protected static final String rst2rst = "JRSTWriter";

    /** XSL Stylesheet to transform Docbook into ODF. */
    protected static final String docbook2odf = "/xsl/docbook2odf-0.244/docbook.xsl";
    /** XSL Stylesheet to transform Docbook into PDF. */
    protected static final String docbook2fo = "/docbook/fo/docbook.xsl";

    public static final String PATTERN_TYPE = "xml|xhtml|docbook|html|xdoc|rst|fo|pdf|odt|rtf";

    /** HTML output format type */
    public static final String TYPE_HTML = "html";

    /** XDOC output format type */
    public static final String TYPE_XDOC = "xdoc";

    /** DOCBOOK output format type */
    public static final String TYPE_DOCBOOK = "docbook";

    /** XHTML output format type */
    public static final String TYPE_XHTML = "xhtml";

    /** JAVA HELP output format type */
    public static final String TYPE_JAVAHELP = "javahelp";

    /** HTML HELP output format type */
    public static final String TYPE_HTMLHELP = "htmlhelp";

    /** RST output format type */
    public static final String TYPE_RST = "rst";

    /** ODT output format type */
    public static final String TYPE_ODT = "odt";

    /** FO output format type */
    public static final String TYPE_FO = "fo";

    /** PDF output format type */
    public static final String TYPE_PDF = "pdf";

    /**
     * key, Out type; value: chain of XSL file to provide wanted file for output
     */
    protected static Map<String, String> stylesheets;

    static {
        stylesheets = new HashMap<String, String>();
        stylesheets.put(TYPE_HTML, rst2xhtml);
        stylesheets.put(TYPE_XDOC, rst2xdoc);
        stylesheets.put(TYPE_DOCBOOK, rst2docbook);
        stylesheets.put(TYPE_XHTML, rst2docbook + "," + docbook2xhtml);
        stylesheets.put(TYPE_JAVAHELP, rst2docbook + "," + docbook2javahelp);
        stylesheets.put(TYPE_HTMLHELP, rst2docbook + "," + docbook2htmlhelp);
        stylesheets.put(TYPE_RST, "");
        stylesheets.put(TYPE_ODT, rst2docbook + "," + docbook2odf);
        stylesheets.put(TYPE_FO, rst2docbook + "," + docbook2fo);
        stylesheets.put(TYPE_PDF, rst2docbook + "," + docbook2fo);
    }

    /**
     * Main method.
     * 
     * @param args main args
     * @throws Exception 
     */
    public static void main(String[] args) throws Exception {
        
        I18n.init(new ClassPathI18nInitializer(), Locale.UK);

        if (args.length == 0) {
            args = askOption();
        }
        if (args == null) {
            System.exit(0);
        }

        ApplicationConfig config = JRSTConfig.getConfig(args);
        config.doAction(0);

        // parse options
        String xslList = JRSTConfigOption.XSL_FILE.getOption(config);
        if (xslList == null) {
            xslList = JRSTConfigOption.OUT_TYPE.getOption(config);
        }
        List<String> unparsed = config.getUnparsed();
        if (unparsed.isEmpty()) {
            JRSTConfig.help();
        }
        File inputFile = new File(config.getUnparsed().get(0));
        File ouputFile = JRSTConfigOption.OUT_FILE.getOptionAsFile(config);
        Overwrite overwrite = Overwrite.NEVER;
        if (JRSTConfigOption.FORCE.getOptionAsBoolean(config)) {
            overwrite = Overwrite.ALLTIME;
        }
        
        generate(xslList, inputFile, ouputFile, overwrite);
    }

    private static String[] askOption() throws SecurityException,
            NoSuchMethodException, IOException {
        String[] result = null;
        try {
            GraphicsEnvironment ge = GraphicsEnvironment
                    .getLocalGraphicsEnvironment();
            GraphicsDevice[] gs = ge.getScreenDevices();
            if (!(gs == null)) {
                if (gs.length > 0) {
                    result = askOptionGraph();
                }
            }
        } catch (java.awt.HeadlessException e) {
            result = null;
        }

        return result;

    }

    /**
     * Interface graphique.
     * 
     * @return
     * @throws SecurityException
     * @throws NoSuchMethodException
     */
    protected static String[] askOptionGraph() throws SecurityException,
            NoSuchMethodException {

        JRSTInterface graph = new JRSTInterface(PATTERN_TYPE);

        return graph.getCmd();
    }

    public static void generate(String xslListOrOutType, File fileIn,
            File fileOut, Overwrite overwrite) throws Exception {
        generate(xslListOrOutType, fileIn, "UTF-8", fileOut, "UTF-8", overwrite);
    }

    /**
     * 
     * @param xslListOrOutType
     * @param fileIn
     * @param inputEncoding
     * @param fileOut
     * @param outputEncoding
     * @param overwrite
     * @throws Exception
     */
    public static void generate(String xslListOrOutType, File fileIn,
            String inputEncoding, File fileOut, String outputEncoding,
            Overwrite overwrite) throws Exception {
        if (fileOut != null
                && fileOut.exists()
                && (overwrite == Overwrite.NEVER || (overwrite == Overwrite.IFNEWER && FileUtil
                        .isNewer(fileIn, fileOut)))) {
            // System.err.println("Don't generate file " + fileOut +
            // ", because already exists");
            log.info("Don't generate file " + fileOut
                    + ", because already exists");
        } else {
            // search xsl file list to apply
            String xslList = stylesheets.get(xslListOrOutType);
            if (xslListOrOutType == null) {
                xslList = xslListOrOutType;
            }

            // parse rst file
            URL url = fileIn.toURI().toURL();
            Reader in = new InputStreamReader(url.openStream(), inputEncoding);
            JRSTReader jrst = new JRSTReader();
            Document doc = jrst.read(in);

            // Sortie vers rst
            if (xslListOrOutType.equals(TYPE_RST)){
                // Creation d'un visitor qui convertie de l'xml vers le rst
                DocUtilsVisitor visitor = new DocUtils2RST();

                // Atacher le visitor au document
                // il va parcourir tout les elements et reconstruire du rst
                doc.accept(visitor);

                // Recuperation du resultat
                String result = visitor.getResult();
                // nettoyage du visiteur
                visitor.clear();
                // Ecriture du resultat dans un fichier
                // prepare the output flux
                FileWriter out = null;
                if (fileOut != null) {
                    try {
                        fileOut.getAbsoluteFile().getParentFile().mkdirs();
                        out = new FileWriter(fileOut);
                        // write generated document
                        out.write(result);
                    } finally {
                        out.close();
                    }
                } else {
                    // Si aucun fichier de sortie nest definie, on utilise la sortie standard
                    System.out.println(result);
                }
            }
            else {
                // apply xsl on rst xml document
                JRSTGenerator gen = new JRSTGenerator();
                String[] xsls = StringUtil.split(xslList, ",");
                for (String xsl : xsls) {
                    URL stylesheet = null;
                    File file = new File(xsl);
                    if (file.exists()) {
                        stylesheet = file.toURI().toURL();
                    } else {
                        //stylesheet = JRST.class.getResource(xsl);
                        stylesheet = Resource.getURL(xsl);
                    }
                    if (stylesheet == null) {
                        throw new FileNotFoundException("Can't find stylesheet: "
                                + xsl);
                    }

                    final String finalXsl = xsl;

                    // add entity resolver
                    gen.setUriResolver(new URIResolver() {
                        public Source resolve(String href, String base) {
                            if (log.isDebugEnabled()) {
                                log.debug("RESOLVING: href: "+href+" base: "+base);
                            }
                            String xslDir = finalXsl.substring(0, finalXsl.lastIndexOf('/') + 1);
                            String resouceFullName = xslDir + href;
                            
                            if (log.isDebugEnabled()) {
                                log.debug("RESOLVING: resouceFullName: "+resouceFullName);
                            }

                            try {
                                InputStream stream = null;
                                File file = new File(resouceFullName);
                                if (file.exists()) {
                                    stream = new FileInputStream(file);
                                } else {
                                    try {
                                        stream = Resource.getURL(resouceFullName).openStream();
                                    }
                                    catch (Exception e) {
                                        stream = Resource.getURL("/docbook/common/" + href).openStream();
                                    }
                                }

                                SAXParserFactory parserFactory = SAXParserFactory.newInstance();  
                                SAXParser saxParser = parserFactory.newSAXParser();  
                                XMLReader xmlReader = saxParser.getXMLReader();
                                xmlReader.setEntityResolver(new EntityResolver() {
                                    @Override
                                    public InputSource resolveEntity(String publicId, String systemId)
                                            throws SAXException, IOException {

                                        InputSource source = null;

                                        String smallSystemId = systemId.substring(systemId.lastIndexOf("/") + 1);
                                        URL url = Resource.getURL("/docbook/common/" + smallSystemId);
                                        if (url != null) {
                                            if (log.isDebugEnabled()) {
                                                log.debug("Resolved entity url : " + url);
                                            }
                                            InputStream stream = url.openStream();
                                            source = new InputSource(stream);
                                        }

                                        return source;
                                    }
                                });
                                InputSource inSrc = new InputSource(stream);
                                SAXSource saxs = new SAXSource(xmlReader, inSrc);
                                return saxs;
                            } catch (IOException e) {
                                e.printStackTrace();
                                return null;
                            } catch (ParserConfigurationException e) {
                                e.printStackTrace();
                                return null;
                            } catch (SAXException e) {
                                e.printStackTrace();
                                return null;
                            }
                        }
                    });

                    doc = gen.transform(doc, stylesheet);
                }

                boolean pdf = false;
                // generation PDF
                if (xslListOrOutType != null) {
                    if (xslListOrOutType.equals("pdf")) {
                        pdf = true;
                        FopFactory fopFactory = FopFactory.newInstance();
                        // OutputStream outPDF = new BufferedOutputStream(new
                        // FileOutputStream(new File("C:/Temp/myfile.pdf")));

                        OutputStream outPDF = null;
                        if (fileOut != null) {
                            fileOut.getAbsoluteFile().getParentFile().mkdirs();
                            outPDF = new BufferedOutputStream(new FileOutputStream(
                                    fileOut));
                        } else {
                            outPDF = new BufferedOutputStream(System.out);
                        }

                        FOUserAgent userAgent = fopFactory.newFOUserAgent();

                        // Step 3: Construct fop with desired output format
                        Fop fop = fopFactory.newFop(MimeConstants.MIME_PDF,
                                userAgent, outPDF);

                        // Step 4: Setup JAXP using identity transformer
                        TransformerFactory factory = TransformerFactory
                                .newInstance();
                        Transformer transformer = factory.newTransformer(); // identity
                        // transformer

                        // Step 5: Setup input and output for XSLT transformation
                        // Setup input stream
                        Source src = new StreamSource(new StringReader(doc.asXML()));

                        // Resulting SAX events (the generated FO) must be piped
                        // through to FOP
                        Result res = new SAXResult(fop.getDefaultHandler());

                        // Step 6: Start XSLT transformation and FOP processing
                        transformer.transform(src, res);

                        if (fileOut != null) {
                            outPDF.close();
                        }
                    }
                }
                if (!pdf) {
                    // prepare the output flux
                    XMLWriter out = null;
                    if (fileOut != null) {
                        fileOut.getAbsoluteFile().getParentFile().mkdirs();

                        out = new XMLWriter(FileUtil.getWriter(fileOut,
                                outputEncoding), new OutputFormat("  ", true,
                                outputEncoding));
                    } else {
                        out = new XMLWriter(System.out, new OutputFormat("  ",
                                true, outputEncoding));
                    }
                    // write generated document
                    out.write(doc);

                    if (fileOut != null) {
                        out.close();
                    }
                }
            }
        }
    }
}
